@@ -23,13 +23,23 @@
  * @since 1.19
  */
 
+use MediaWiki\Api\ApiQueryBase;
+use MediaWiki\Api\ApiResult;
+use MediaWiki\CommentFormatter\CommentFormatter;
+use MediaWiki\Context\IContextSource;
+use MediaWiki\Context\RequestContext;
 use MediaWiki\Html\Html;
+use MediaWiki\Language\Language;
 use MediaWiki\Linker\Linker;
 use MediaWiki\Linker\LinkRenderer;
 use MediaWiki\Linker\LinkTarget;
 use MediaWiki\MainConfigNames;
 use MediaWiki\MediaWikiServices;
+use MediaWiki\Message\Message;
+use MediaWiki\SpecialPage\SpecialPage;
 use MediaWiki\Title\Title;
+use MediaWiki\User\User;
+use MediaWiki\User\UserEditTracker;
 use MediaWiki\User\UserIdentity;
 
 /**
@@ -55,19 +65,11 @@ class LogFormatter {
 	 * Constructs a new formatter suitable for given entry.
 	 * @param LogEntry $entry
 	 * @return LogFormatter
+	 * @deprecated since 1.42, use LogFormatterFactory instead, hard-deprecated since 1.43
 	 */
 	public static function newFromEntry( LogEntry $entry ) {
-		$logActionsHandlers = MediaWikiServices::getInstance()->getMainConfig()
-			->get( MainConfigNames::LogActionsHandlers );
-		$fulltype = $entry->getFullType();
-		$wildcard = $entry->getType() . '/*';
-		$handler = $logActionsHandlers[$fulltype] ?? $logActionsHandlers[$wildcard] ?? '';
-
-		if ( $handler !== '' && is_string( $handler ) && class_exists( $handler ) ) {
-			return new $handler( $entry );
-		}
-
-		return new LegacyLogFormatter( $entry );
+		wfDeprecated( __METHOD__, '1.42' );
+		return MediaWikiServices::getInstance()->getLogFormatterFactory()->newFromEntry( $entry );
 	}
 
 	/**
@@ -76,8 +78,10 @@ class LogFormatter {
 	 * @param stdClass|array $row
 	 * @see DatabaseLogEntry::getSelectQueryData
 	 * @return LogFormatter
+	 * @deprecated since 1.42, use LogFormatterFactory instead, hard-deprecated since 1.43
 	 */
 	public static function newFromRow( $row ) {
+		wfDeprecated( __METHOD__, '1.42' );
 		return self::newFromEntry( DatabaseLogEntry::newFromRow( $row ) );
 	}
 
@@ -107,11 +111,18 @@ class LogFormatter {
 	/** @var bool */
 	protected $irctext = false;
 
-	/**
-	 * @var LinkRenderer|null
-	 */
+	/** @var LinkRenderer|null */
 	private $linkRenderer;
 
+	/** @var Language|null */
+	private $contentLanguage;
+
+	/** @var CommentFormatter|null */
+	private $commentFormatter;
+
+	/** @var UserEditTracker|null */
+	private $userEditTracker;
+
 	/**
 	 * @see LogFormatter::getMessageParameters
 	 * @var array
@@ -123,7 +134,7 @@ class LogFormatter {
 	 *
 	 * @param LogEntry $entry
 	 */
-	protected function __construct( LogEntry $entry ) {
+	public function __construct( LogEntry $entry ) {
 		$this->entry = $entry;
 		$this->context = RequestContext::getMain();
 	}
@@ -152,10 +163,74 @@ class LogFormatter {
 		if ( $this->linkRenderer !== null ) {
 			return $this->linkRenderer;
 		} else {
+			wfDeprecated( static::class . " without all required services", '1.42' );
 			return MediaWikiServices::getInstance()->getLinkRenderer();
 		}
 	}
 
+	/**
+	 * @internal For factory only
+	 * @since 1.42
+	 * @param Language $contentLanguage
+	 */
+	final public function setContentLanguage( Language $contentLanguage ) {
+		$this->contentLanguage = $contentLanguage;
+	}
+
+	/**
+	 * @since 1.42
+	 * @return Language
+	 */
+	final public function getContentLanguage(): Language {
+		if ( $this->contentLanguage === null ) {
+			wfDeprecated( static::class . " without all required services", '1.42' );
+			$this->contentLanguage = MediaWikiServices::getInstance()->getContentLanguage();
+		}
+		return $this->contentLanguage;
+	}
+
+	/**
+	 * @internal For factory only
+	 * @since 1.42
+	 * @param CommentFormatter $commentFormatter
+	 */
+	final public function setCommentFormatter( CommentFormatter $commentFormatter ) {
+		$this->commentFormatter = $commentFormatter;
+	}
+
+	/**
+	 * @since 1.42
+	 * @return CommentFormatter
+	 */
+	final public function getCommentFormatter(): CommentFormatter {
+		if ( $this->commentFormatter === null ) {
+			wfDeprecated( static::class . " without all required services", '1.42' );
+			$this->commentFormatter = MediaWikiServices::getInstance()->getCommentFormatter();
+		}
+		return $this->commentFormatter;
+	}
+
+	/**
+	 * @internal For factory only
+	 * @since 1.42
+	 * @param UserEditTracker $userEditTracker
+	 */
+	final public function setUserEditTracker( UserEditTracker $userEditTracker ) {
+		$this->userEditTracker = $userEditTracker;
+	}
+
+	/**
+	 * @since 1.42
+	 * @return UserEditTracker
+	 */
+	final public function getUserEditTracker(): UserEditTracker {
+		if ( $this->userEditTracker === null ) {
+			wfDeprecated( static::class . " without all required services", '1.42' );
+			$this->userEditTracker = MediaWikiServices::getInstance()->getUserEditTracker();
+		}
+		return $this->userEditTracker;
+	}
+
 	/**
 	 * Set the visibility restrictions for displaying content.
 	 * If set to public, and an item is deleted, then it will be replaced
@@ -260,7 +335,7 @@ class LogFormatter {
 		// Text of title the action is aimed at.
 		$target = $entry->getTarget()->getPrefixedText();
 		$text = null;
-		$contLang = MediaWikiServices::getInstance()->getContentLanguage();
+		$contLang = $this->getContentLanguage();
 		switch ( $entry->getType() ) {
 			case 'move':
 				switch ( $entry->getSubtype() ) {
@@ -565,14 +640,21 @@ class LogFormatter {
 	}
 
 	/**
-	 * Formats parameters intended for action message from
-	 * array of all parameters. There are three hardcoded
-	 * parameters (array is zero-indexed, this list not):
-	 *  - 1: user name with premade link
-	 *  - 2: usable for gender magic function
-	 *  - 3: target page with premade link
+	 * Formats parameters intended for action message from array of all parameters.
+	 * There are three hardcoded parameters:
+	 *  - $1: user name with premade link
+	 *  - $2: usable for gender magic function
+	 *  - $3: target page with premade link
+	 * More parameters might be present, depending on what code created the log
+	 * entry.
+	 *
+	 * The parameters are returned as a non-associative array that can be passed to
+	 * Message::params(), so $logFormatter->getMessageParameters()[0] is the $1 parameter
+	 * in the message and so on.
+	 *
 	 * @stable to override
 	 * @return array
+	 * @see ManualLogEntry::setParameters() for how parameters are determined.
 	 */
 	protected function getMessageParameters() {
 		if ( isset( $this->parsedParameters ) ) {
@@ -615,7 +697,7 @@ class LogFormatter {
 	 *     * number: Format value as number
 	 *     * list: Format value as a comma-separated list
 	 * @param mixed $value The parameter value that should be formatted
-	 * @return string|array Formatted value
+	 * @return mixed Formatted value
 	 * @since 1.21
 	 */
 	protected function formatParameterValue( $type, $value ) {
@@ -680,7 +762,7 @@ class LogFormatter {
 	 * @return string wikitext or html
 	 * @return-taint onlysafefor_html
 	 */
-	protected function makePageLink( Title $title = null, $parameters = [], $html = null ) {
+	protected function makePageLink( ?Title $title = null, $parameters = [], $html = null ) {
 		if ( !$title instanceof Title ) {
 			$msg = $this->msg( 'invalidtitle' )->text();
 			if ( $this->plaintext ) {
@@ -727,7 +809,7 @@ class LogFormatter {
 	 */
 	public function getComment() {
 		if ( $this->canView( LogPage::DELETED_COMMENT ) ) {
-			$comment = MediaWikiServices::getInstance()->getCommentFormatter()
+			$comment = $this->getCommentFormatter()
 				->formatBlock( $this->entry->getComment() );
 			// No hard coded spaces thanx
 			$element = ltrim( $comment );
@@ -797,9 +879,7 @@ class LogFormatter {
 				$user->getName()
 			);
 			if ( $this->linkFlood ) {
-				$editCount = $user->isRegistered()
-					? MediaWikiServices::getInstance()->getUserEditTracker()->getUserEditCount( $user )
-					: null;
+				$editCount = $this->getUserEditTracker()->getUserEditCount( $user );
 
 				$element .= Linker::userToolLinks(
 					$user->getId(),
